/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 2 -*-
 *
 *  Copyright 2005  Carlos Garnacho Parro
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of version 2 of the GNU General Public
 *  License as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 *  Authors: Carlos Garnacho Parro  <carlosg@gnome.org>
 *
 */

#include <gtk/gtk.h>
#include <mail/em-popup.h>
#include <libedataserver/md5-utils.h>
#include <e-util/e-error.h>

#include <camel/camel-folder.h>
#include <camel/camel-medium.h>
#include <camel/camel-stream.h>
#include <camel/camel-stream-mem.h>

static gint
run_dialog (gint n_duplicates)
{
  gchar *str;
  gint response;

  str = g_strdup_printf ("%i", n_duplicates);
  response = e_error_run (NULL, "org.gnome.remove-duplicates:delete-duplicates", str, NULL);
  g_free (str);

  return response;
}

static void
delete_message_foreach (gpointer value, gpointer data)
{
  gchar *uid = value;
  CamelFolder *folder = data;

  camel_folder_delete_message (folder, uid);
}

static guchar*
get_message_md5 (CamelFolder *folder,
		 const gchar *uid)
{
  CamelMimeMessage *msg;
  CamelException    ex;
  CamelDataWrapper *content;
  CamelStream      *mem;
  guchar           *digest = NULL;

  camel_exception_init (&ex);
  msg = camel_folder_get_message (folder, uid, &ex);

  if (camel_exception_is_set (&ex))
    {
      camel_exception_clear (&ex);
      return NULL;
    }

  /* get message contents */
  content = camel_medium_get_content_object ((CamelMedium *) msg);

  if (!content)
    return NULL;

  /* calculate MD5 */
  mem = camel_stream_mem_new ();
  camel_data_wrapper_decode_to_stream (content, mem);
  digest = g_new0 (guchar, 16);

  md5_get_digest (((CamelStreamMem *) mem)->buffer->data,
		  ((CamelStreamMem *) mem)->buffer->len,
		  digest);

  camel_object_unref (mem);
  camel_object_unref (msg);

  return digest;
}

static gboolean
message_is_duplicated (GHashTable *messages, guint64 id, guchar *digest)
{
  guchar *hash_digest;
  int i;

  hash_digest = g_hash_table_lookup (messages, &id);

  if (!hash_digest)
    return FALSE;

  for (i = 0; i < 16; i++)
    if (digest[i] != hash_digest[i])
      return FALSE;

  return TRUE;
}

static void
handle_messages (CamelFolder *folder, GPtrArray *uids)
{
  CamelMessageInfo *msg_info;
  const CamelSummaryMessageID *mid;
  GHashTable *messages;
  GSList *duplicates = NULL;
  guint32 flags;
  guchar *digest;
  gint i, n_duplicates;

  messages = g_hash_table_new_full (g_int_hash, g_int_equal, g_free, g_free);

  for (i = 0; i < uids->len; i++)
    {
      msg_info = camel_folder_get_message_info (folder, uids->pdata[i]);
      digest   = get_message_md5 (folder, uids->pdata[i]);
      mid      = camel_message_info_message_id (msg_info);
      flags    = camel_message_info_flags (msg_info);

      if (!(flags & CAMEL_MESSAGE_DELETED))
        {
	  if (message_is_duplicated (messages, mid->id.id, digest))
	    duplicates = g_slist_prepend (duplicates, g_strdup (uids->pdata[i]));
	  else
	    {
	      guint64 *id;

	      id = g_new0 (guint64, 1);
	      *id = mid->id.id;
	      g_hash_table_insert (messages, id, digest);
	    }
	}

      camel_message_info_free (msg_info);
    }

  n_duplicates = g_slist_length (duplicates);

  if (n_duplicates && (run_dialog (n_duplicates) == GTK_RESPONSE_OK))
    g_slist_foreach (duplicates, delete_message_foreach, folder);

  g_hash_table_destroy (messages);
  g_slist_foreach (duplicates, (GFunc) g_free, NULL);
  g_slist_free (duplicates);
}

void org_gnome_duplicates_remove (EPlugin *ep, EMPopupTargetSelect *target);

void
org_gnome_duplicates_remove (EPlugin *ep, EMPopupTargetSelect *target)
{
  handle_messages (target->folder, target->uids);
}
